#!/usr/bin/env python3 

import numpy as np
import networkx as nx
from ase.autoneb import AutoNEB
from ase.constraints import ExternalForce,FixAtoms
from ase.dimer import DimerControl, MinModeAtoms, MinModeTranslate
from ase.io import read, write
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution
from ase.optimize import BFGS,FIRE
from ase.vibrations import Vibrations
#from ase.build import minimize_rotation_and_translation
from ase import units
from createMat import get_G_index
from mopacamk import MOPACamk
from os import system
from shutil import copyfile
from re import search
from sys import argv, exit
from Integrators import Langevin
from xtb.ase.calculator import XTB
from AMK_parameters import emax, label, task
from subprocess import run,PIPE
from sella import Sella
import networkx.algorithms.isomorphism as iso


def rotation_matrix_from_points(m0, m1):
    """Returns a rigid transformation/rotation matrix that minimizes the
    RMSD between two set of points.

    m0 and m1 should be (3, npoints) numpy arrays with
    coordinates as columns::

        (x1  x2   x3   ... xN
         y1  y2   y3   ... yN
         z1  z2   z3   ... zN)

    The centeroids should be set to origin prior to
    computing the rotation matrix.

    The rotation matrix is computed using quaternion
    algebra as detailed in::

        Melander et al. J. Chem. Theory Comput., 2015, 11,1055
    """

    v0 = np.copy(m0)
    v1 = np.copy(m1)

    # compute the rotation quaternion

    R11, R22, R33 = np.sum(v0 * v1, axis=1)
    R12, R23, R31 = np.sum(v0 * np.roll(v1, -1, axis=0), axis=1)
    R13, R21, R32 = np.sum(v0 * np.roll(v1, -2, axis=0), axis=1)

    f = [[R11 + R22 + R33, R23 - R32, R31 - R13, R12 - R21],
         [R23 - R32, R11 - R22 - R33, R12 + R21, R13 + R31],
         [R31 - R13, R12 + R21, -R11 + R22 - R33, R23 + R32],
         [R12 - R21, R13 + R31, R23 + R32, -R11 - R22 + R33]]

    F = np.array(f)

    w, V = np.linalg.eigh(F)
    # eigenvector corresponding to the most
    # positive eigenvalue
    q = V[:, np.argmax(w)]

    # Rotation matrix from the quaternion q

    R = quaternion_to_matrix(q)

    return R

def quaternion_to_matrix(q):
    """Returns a rotation matrix.

    Computed from a unit quaternion Input as (4,) numpy array.
    """

    q0, q1, q2, q3 = q
    R_q = [[q0**2 + q1**2 - q2**2 - q3**2,
            2 * (q1 * q2 - q0 * q3),
            2 * (q1 * q3 + q0 * q2)],
           [2 * (q1 * q2 + q0 * q3),
            q0**2 - q1**2 + q2**2 - q3**2,
            2 * (q2 * q3 - q0 * q1)],
           [2 * (q1 * q3 - q0 * q2),
            2 * (q2 * q3 + q0 * q1),
            q0**2 - q1**2 - q2**2 + q3**2]]
    return np.array(R_q)


def minimize_rotation_and_translation(target, atoms, weight,wa):
    """Minimize RMSD between atoms and target.

    Rotate and translate atoms to best match target.  For more details, see::

        Melander et al. J. Chem. Theory Comput., 2015, 11,1055
    """

    p = atoms.get_positions()
    p0 = target.get_positions()

    # centeroids to origin
    c = np.mean(p, axis=0)
    p -= c
    c0 = np.mean(p0, axis=0)
    p0 -= c0

    #EMN
    for x in wa:
       p0[x] *= weight
       p[x]  *= weight
    #EMN

    # Compute rotation matrix
    R = rotation_matrix_from_points(p.T, p0.T)


    #EMN
    for x in wa:
       p0[x] /= weight
       p[x]  /= weight
    #EMN

    atoms.set_positions(np.dot(p, R.T) + c0)


def distort(atomos,mol):
    v1 = mol.get_positions()[atomos[2]] - mol.get_positions()[atomos[1]]
    v2 = mol.get_positions()[atomos[0]] - mol.get_positions()[atomos[2]]
    v3 = mol.get_positions()[atomos[1]] - mol.get_positions()[atomos[0]]
    nv1 = np.linalg.norm(v1) ; nv2 = np.linalg.norm(v2) ; nv3 = np.linalg.norm(v3)
    if abs( round( np.dot(v1 / nv1,v2 / nv2) , 2)) == 1.0:
       l = [nv1,nv2,nv3] ; ml = max(l) ; middle = l.index(ml)
       atd = atomos[middle]
       posit = mol.get_positions()
       a = (v1[1] / v1[0]) ** 2 + 1
       b = v1[2] * v1[1]/ v1[0] ** 2
       c = (v1[2] / (2 * v1[0]) ) ** 2 - 3 / 4
       uy = (- b + np.sqrt( b ** 2 - 4 * a * c ) ) / (2 * a)
       ux = np.sqrt(3 / 4 - uy * uy)
       uz = 0.5
       u = np.array([ux,uy,uz])
       posit[atd] = posit[atd] +  0.8 * u
       mol.set_positions(posit)
       return True
    else: return False

def vib_calc(ts):
    vib = Vibrations(ts)
    vib.run(); vib.summary(); vib.clean()
    freq = vib.get_frequencies()
    eigenv = []
    for i in range(len(freq)):
        eigenv.append(vib.get_mode(i))
    return eigenv,freq

def attach_calculators(images):
    for image in images:  
        if prog == 'mopac': image.calc = MOPACamk(method=method,relscf=1.0)
        elif prog == 'XTB': image.calc = XTB(method=method)

def Energy_and_forces(geom):
    # Get forces and energy from designated potential
    try: ene = geom.get_potential_energy()
    except: exit()
    forces = geom.get_forces()
    return ene,forces

def runTrajectory(geom, T, Fric, totaltime, timeStep,breakl,forml):
    global n_images
    #n_dist = int( len(geom) * (len(geom) - 1) / 2)
    ene,forces = Energy_and_forces(geom) 
    mdInt = Langevin(units.kB * T, Fric, forces, geom.get_velocities(), geom, timeStep)
    ene,forces = Energy_and_forces(geom)

    # Run MD trajectory for specified number of steps
    n_stop_criteria = 0; damped = False
    write("traj.xyz",geom.copy(),format="xyz")
    for i in range(0,int(totaltime/dt)):
        ene,forces = Energy_and_forces(geom)
        mdInt.mdStepPos(forces,timeStep,geom)
        ene,forces = Energy_and_forces(geom)
        mdInt.mdStepVel(forces,timeStep,geom,damped)

        #Print current positions to file
        write("traj.xyz",geom.copy(),format="xyz",append=True)
        #check adjacency matrix
        G,ind,jnd,ibn,jbn = get_G_index(geom,1,len(geom),False)
        for n in range(n_dist):
            if G[ind[n]][jnd[n]]['weight'] == 0: G.remove_edge(ind[n],jnd[n])
        criteria = 0
        for ele in breakl: criteria += G.has_edge(ele[0],ele[1])
        for ele in forml:  
            criteria += not G.has_edge(ele[0],ele[1])
            if geom.get_distance(ele[0],ele[1]) < 1: 
                damped = True ; geom.set_constraint()
        if criteria == 0: n_stop_criteria += 1
        if n_stop_criteria >= 6: return G
        elif n_stop_criteria >= 3: geom.set_constraint()
        elif n_stop_criteria == 1: damped = True
    geom.set_constraint()
    return G
             
inputfile = str(argv[1]) ; line = int(argv[2])
system('rm -rf image*.traj')
#Default parameters
n_max,prefix,fmax,temp,fric,totaltime,dt,method,prog,ExtForce,weight = 10,'image',0.1,0.,0.5,100,1,'GFN2-xTB','XTB',6,100
#Here we should read inputfile
for linei in open(inputfile,'r'):
    if search("LowLevel ", linei): prog   = str(linei.split()[1]) 
    if search("LowLevel ", linei): method = str(linei.split()[2]) 
    if search("molecule ", linei): molecule = str(linei.split()[1])
    if search("emax ", linei): emax = float(linei.split()[1]) 
    if search("ExtForce ", linei): ExtForce = float(linei.split()[1]) 
    if search("tsdirll ", linei): 
        path = str(linei.split()[1]) 
try: 
    print('Path to files:',path)
except:
    path = 'tsdirLL_'+molecule
    print('Path to files:',path)

#We now read ts_bonds.inp file
tsbfile = open(path+'/ts_bonds.inp', "r")
lines = tsbfile.readlines()
cs = lines[line].split() ; constr = []; breakl = []; forml = []; atoms_rxn = []
tsbfile.close()
for i in range(0,len(cs),3):
    if int(cs[i+1]) not in atoms_rxn: atoms_rxn.append(int(cs[i+1]))
    if int(cs[i+2]) not in atoms_rxn: atoms_rxn.append(int(cs[i+2]))
    if cs[i] == "b":
        c = ExternalForce( int(cs[i+1]) , int(cs[i+2]) ,ExtForce)
        constr.append(c) ; breakl.append([ int(cs[i+1]) , int(cs[i+2]) ])
    elif cs[i] == "f":
        c = ExternalForce( int(cs[i+1]) , int(cs[i+2]) ,-ExtForce)
        constr.append(c) ; forml.append([ int(cs[i+1]) , int(cs[i+2]) ])
#Instantiate rmol
rmol = read(molecule+'.xyz')
n_dist = int( len(rmol) * (len(rmol) - 1) / 2)
natom = len(rmol)
aton   = rmol.get_atomic_numbers()
symb   = rmol.get_chemical_symbols()
if prog == 'mopac': rmol.calc = MOPACamk(method=method+' threads=1',relscf=1.0)
elif prog == 'XTB': rmol.calc = XTB(method=method)
#atoms_not_rxn
latoms = [item for item in range(natom)]
atoms_not_rxn = np.setdiff1d(latoms,atoms_rxn)

#Optimization of the molecule
print('Path:',line,cs)
print('')
print('Optimizing reactant...')
opt = BFGS(rmol, trajectory='image000.traj',logfile='bfgs000.log')
opt.run(fmax=fmax)
react = rmol.copy()
write("react.xyz",react)
e0 = rmol.get_potential_energy()
if prog == 'XTB':
    reactfile = open(molecule+'_freq.out', 'w')
    reactfile.write('Energy= '+str(e0)+'\n')
print('Reactant optimized')


##################G of the reactant
Gr,ind,jnd,ibn,jbn = get_G_index(rmol,1,len(rmol),False)
for n in range(n_dist):
    if Gr[ind[n]][jnd[n]]['weight'] == 0: Gr.remove_edge(ind[n],jnd[n])
A  = nx.adjacency_matrix(Gr) ; Ar = A.A
for z in range(natom): Ar[z][z] = aton[z]
##################G of the expected product
Gp = Gr.copy()
for ele in breakl: Gp.remove_edge(ele[0],ele[1])
for ele in forml:  Gp.add_edge(ele[0],ele[1])
A  = nx.adjacency_matrix(Gp) ; Ap = A.A
for z in range(natom): Ap[z][z] = aton[z]
##################
tag_p = np.array(sorted( [np.round(elem,3) for elem in np.linalg.eigvals(Ap) ] ))
tag_r = np.array(sorted( [np.round(elem,3) for elem in np.linalg.eigvals(Ar) ] ))
if np.linalg.norm(tag_p - tag_r) == 0: weight = 1

#When three atoms involved in the forces are in a line--> distort the geometry and add one more image
n_images = 1 
if len(atoms_rxn) == 3:
    if distort(atoms_rxn,rmol): 
        print('Reactant geometry distorted')
        minimize_rotation_and_translation(react,rmol,weight,atoms_not_rxn)
        write('react_distorted.xyz',rmol)
        write('image00'+str(n_images)+'.traj',rmol)
        n_images += 1
        rmol.set_constraint() 
###

#For the dynamics we give all Hs a mass of 4.0 and apply contraints
masses = []
for x in aton:
    if x == 1: masses.append(4.0)
    else: masses.append(None)
rmol.set_masses(masses=masses) ; rmol.set_constraint(constr)

###
MaxwellBoltzmannDistribution(rmol, temp * units.kB)
##Run a constrained short traj to reach the prod.
G = runTrajectory(rmol,temp,fric,totaltime,dt * units.fs ,breakl,forml)

print('Optimizing product...')
minimize_rotation_and_translation(react,rmol,weight,atoms_not_rxn)
#For intermediates we first move the structure along the largest negative eigenvector
if nx.is_connected(Gp):
    eigenv, freq = vib_calc(rmol)
    #We first move the structure along the largest negative eigenvector direction 
    positions = rmol.get_positions() + 0.01 * eigenv[0] / np.linalg.norm(eigenv[0])
    rmol.set_positions(positions)
    #print('First using GFNFF...')
    #rmol.calc = XTB(method='GFNFF') 
    #opt = BFGS(rmol, trajectory='gfnff.traj',logfile='bfgs_gfnff.log')
    #opt.run(fmax=fmax)
    ##Switch back to the semiempirical
    #if prog == 'mopac': rmol.calc = MOPACamk(method=method+' threads=1',relscf=1.0)
    #elif prog == 'XTB': rmol.calc = XTB(method=method)
opt = BFGS(rmol, trajectory='image00'+str(n_images)+'.traj',logfile='bfgs00'+str(n_images)+'.log')
opt.run(fmax=fmax)
prod = rmol.copy()
write("prod.xyz",prod)
print('Product optimized')

###Gx is the Graph coming out of the optimization 
Gx,ind,jnd,ibn,jbn = get_G_index(rmol,1,len(rmol),False)
for n in range(n_dist):
    if Gx[ind[n]][jnd[n]]['weight'] == 0: Gx.remove_edge(ind[n],jnd[n])
A  = nx.adjacency_matrix(Gx) ; Ax = A.A
for z in range(natom): Ax[z][z] = aton[z]
###Check for barrierless processes
Adiff = Ar - Ax
if np.linalg.norm(Adiff) == 0:
    if not nx.is_connected(Gp): print('Final and initial states are the same --> Barrierless process')
    else: print('It seems that the value of ExtForce is not enought to break/form bonds')
    exit()
###Check that the Gx is isomorphic with Gp
criteria = 0
for ele in breakl: criteria += Gx.has_edge(ele[0],ele[1])
for ele in forml:  criteria += not Gx.has_edge(ele[0],ele[1])
if criteria > 0:
    print('Obtained product is not the expected  --> The product could not be generated')
    exit()
#Adiff = Ap - Ax
#if np.linalg.norm(Adiff) > 0:
#    print('Obtained product is not the expected  --> The product could not be generated')
#    exit()
##################
 
ep = rmol.get_potential_energy()
dE = ep - e0
print('{:s} {:10.4f} {:s}  {:10.2f} {:s}'.format('Product energy: ',dE,'eV', dE * units.mol / units.kcal,'kcal/mol'))


#if nx.is_connected(G):
dE = dE * units.mol / units.kcal
if dE > emax:
    print('Product energy > emax:',dE,emax)
    exit()

print('Dynamics with External Force finished')
print('')
#Run autoneb 
autoneb = AutoNEB(attach_calculators,
                  prefix=prefix, 
                  optimizer='BFGS',
                  n_simul=1,
                  n_max=n_max,
                  fmax=fmax,
                  k=0.1,
                  parallel=False,
                  maxsteps=[50,1000])
try: 
    autoneb.run()
except Exception as e:
    print(e)
    print('ERROR in autoneb calculation')
    exit()

#Get max value along the NEB
gapmax = 0
print('')
print('#   Erel(kcal/mol)        E(eV)')
for i in range(n_max):
    pot = autoneb.all_images[i].get_potential_energy()
    gap = (pot - e0) * units.mol / units.kcal
    if gap > gapmax and pot !=0 and i != n_max-1 and gap < 1.2 * emax: 
        gapmax = gap; imax = i
    print('{:1.0f} {:16.2f} {:12.4f}'.format(i,gap, pot))
print('selected image',imax)

#TS initial guess is the maximum along the NEB
ts = autoneb.all_images[imax].copy()
write('ts_inp.xyz',ts)

#TS optimization
if prog == 'mopac':
    # Use mopac TS optimizer
    ts.calc = MOPACamk(relscf = 1.0,label = 'ts',task = 'ts threads=1 charge=0 precise cycles=1000 t=500 ddmax=0.1 denout',freq=True)
    try: 
        print('{:s} {:10.4f}'.format('TS optimized energy:',ts.get_potential_energy()))
        print('Lowest vibrational frequencies:',[float(x) for x in ts.calc.get_freqs()])
        p = run("check_ts_structure.sh > ts.log",shell=True)
        print(p)
    except Exception as e: 
        print('ERROR in MOPAC calculation')
        exit()
###############################
elif prog == 'XTB':
    #Dimer method for XTB (no internal optimizer)
    #vib calc. to get the lowest frequency mode
    ts.calc = XTB(method=method)
    eigenv, freq = vib_calc(ts)
    lfm0 = eigenv[0] ; lfm1 = eigenv[1]
    print(lfm0)
    print(lfm1)
    print(freq)
    #We first move the ts structure in the second negative eigenvector direction to avoid second order saddles
    positions = ts.get_positions() + 0.01 * lfm1 / np.linalg.norm(lfm1)
    ts.set_positions(positions)

    #set up the dimer calc
    d_control = DimerControl(initial_eigenmode_method = 'displacement', \
            displacement_method = 'vector', logfile = None, mask=[True]*len(rmol))

    d_atoms = MinModeAtoms(ts, d_control)

    displacement_vector =  0.1 * lfm0 / np.linalg.norm(lfm0)
    d_atoms.displace(displacement_vector = displacement_vector)

    dim_rlx=MinModeTranslate(d_atoms, trajectory='dimer_method_traj', logfile=None)
    try:
        dim_rlx.run(fmax=0.001,steps=1000)
    except Exception as e: 
        print('ERROR in dimer calculation')
        exit()

    try:
        eigenv,freq = vib_calc(ts)
        ets = ts.get_potential_energy()
        print('TS optimized energy :',ets)
        tsfile  = open('ts.out', 'w')
        moldenfile  = open('ts.molden', 'w')
        moldenfile.write('[Molden Format]'+'\n\n')
        tsfile.write('Energy= '+str(ets)+'\n')
        tsfile.write('Freq:'+'\n')
        moldenfile.write('[FREQ]'+'\n')
        for i,x in enumerate(freq):
            if i == 0: tsfile.write(str(-x.imag)+'\n')
            elif i >6: tsfile.write(str(x.real)+'\n')
            if i == 0: moldenfile.write(str(-x.imag)+'\n')
            elif i >6: moldenfile.write(str(x.real)+'\n')
        tsfile.write('Gibbs free energy: [0.]'+'\n')
        tsfile.write('ZPE: [0.]'+'\n')
        tsfile.write(str(natom)+'\nFinal structure:'+'\n')
        moldenfile.write('\n[FR-COORD]'+'\n')
        posit = ts.get_positions() 
        for i,ele in enumerate(symb):
            tsfile.write(str(ele)+' '+str(posit[i][0])+' '+str(posit[i][1])+' '+str(posit[i][2])+'\n')
            moldenfile.write(str(ele)+' '+str(posit[i][0]/units.Bohr)+' '+str(posit[i][1]/units.Bohr)+' '+str(posit[i][2]/units.Bohr)+'\n')
        moldenfile.write('\n\n[FR-NORM-COORD]'+'\n')
        for i in range(len(freq)-6):
            moldenfile.write('vibration '+str(i+1)+'\n')
            if i == 0: ifreq = i
            else: ifreq = i + 6
            for j in range(natom):
                moldenfile.write(str(eigenv[ifreq][j][0]/units.Bohr)+' '+str(eigenv[ifreq][j][1]/units.Bohr)+' '+str(eigenv[ifreq][j][2]/units.Bohr)+'\n')
        tsfile.close()
        moldenfile.close()
        write('ts_opt.xyz',ts)
        p = run("check_ts_structure.sh > ts.log",shell=True)
    except Exception as e: 
        print('ERROR in the calculation')
        exit()
  
